/* Copyright (C) 2009 Gabor Kovesdan <gabor@FreeBSD.org>
 * 2012 Oleg Moskalenko oleg.moskalenko@citrix.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/limits.h>

#include <errno.h>
#include <err.h>
#include <langinfo.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <wctype.h>
#include <ctype.h>

#include "bwstring.h"

bool byte_sort=false;

/* counterparts of wcs functions */

void bwsprintf(FILE* f, bwstring* bws, const char* prefix, const char* suffix)
{
	if (MB_CUR_MAX == 1) {
		fprintf(f, "%s%s%s", prefix, bws->data.cstr, suffix);
	} else {
		fprintf(f, "%s%S%s", prefix, bws->data.wstr, suffix);
	}
}

size_t bws_memsize(const bwstring* bws)
{
	if (MB_CUR_MAX == 1) {
		return (bws->sz + sizeof(bwstring));
	} else {
		return (SIZEOF_WCHAR_STRING(bws->sz) + sizeof(bwstring));
	}
}

void bws_setlen(bwstring* bws, size_t newlen)
{
	if (bws && newlen != bws->len && newlen <= bws->sz) {
		bws->len = newlen;
		if (MB_CUR_MAX == 1) {
			bws->data.cstr[newlen] = '\0';
		} else {
			bws->data.wstr[newlen] = L'\0';
		}
	}
}

/*
 * Allocate a new binary string of specified size
 */
bwstring* bwsalloc(size_t sz)
{
	bwstring *ret = NULL;

	if (MB_CUR_MAX == 1) {
		ret = (bwstring*) sort_malloc(sizeof(bwstring) + sz);
	} else {
		ret = (bwstring*) sort_malloc(sizeof(bwstring) + SIZEOF_WCHAR_STRING(sz));
	}

	if (!ret)
		err(2, NULL);
	else {
		ret->sz = sz;
		ret->len = 0;
		if (MB_CUR_MAX == 1) {
			ret->data.cstr[ret->len] = '\0';
		} else {
			ret->data.wstr[ret->len] = L'\0';
		}

		return ret;
	}
}

/*
 * Create a copy of binary string.
 * New string size equals the length of the old string.
 */
bwstring* bwsdup(const bwstring *s)
{
	if (!s)
		return NULL;
	else {
		bwstring* ret = bwsalloc(s->len);
		ret->len = s->len;
		if (MB_CUR_MAX == 1) {
			memcpy(ret->data.cstr, s->data.cstr, (s->len));
			ret->data.cstr[ret->len] = '\0';
		} else {
			memcpy(ret->data.wstr, s->data.wstr, SIZEOF_WCHAR_STRING(s->len));
			ret->data.wstr[ret->len] = L'\0';
		}
		return ret;
	}
}

/*
 * Create a new binary string from a raw binary buffer.
 */
bwstring* bwssbdup(const wchar_t *str, size_t len)
{
	if (!str) {
		if (len == 0) {
			bwstring* ret = bwsalloc(0);
			ret->len = 0;
			if (MB_CUR_MAX == 1) {
				ret->data.cstr[ret->len] = '\0';
			} else {
				ret->data.wstr[ret->len] = L'\0';
			}
			return ret;
		}
		return NULL;
	} else {
		bwstring* ret = bwsalloc(len);
		ret->len = len;
		if (MB_CUR_MAX == 1) {
			size_t i = 0;
			for (i = 0; i < len; ++i) {
				ret->data.cstr[i] = (unsigned char) str[i];
			}
			ret->data.cstr[ret->len] = '\0';
		} else {
			memcpy(ret->data.wstr, str, SIZEOF_WCHAR_STRING(len));
			ret->data.wstr[ret->len] = L'\0';
		}
		return ret;
	}
}

/*
 * Create a new binary string from a raw binary buffer.
 */
bwstring* bwscsbdup(const unsigned char *str, size_t len)
{
	bwstring* ret = bwsalloc(len);
	if (str) {
		if (MB_CUR_MAX == 1) {
			memcpy(ret->data.cstr, str, len);
			ret->len = len;
			ret->data.cstr[ret->len] = '\0';
		} else {
			size_t charlen = 0, chars = 0;
			mbstate_t mbs;
			size_t cptr = 0;
			const char* s = (const char*) str;

			memset(&mbs, 0, sizeof(mbs));

			while (cptr < len) {
				size_t n = MB_CUR_MAX;
				if (n > len - cptr)
					n = len - cptr;
				charlen = mbrlen(s + cptr, n, &mbs);
				switch (charlen){
				case 0:
				case (size_t) -1:
				case (size_t) -2:
					ret->data.wstr[chars++] = (unsigned char) s[cptr];
					++cptr;
					break;
				default:
					mbrtowc(ret->data.wstr + (chars++), s + cptr, charlen, &mbs);
					cptr += charlen;
				};
			}

			ret->len = chars;
			ret->data.wstr[ret->len] = L'\0';
		}
	}
	return ret;
}

/*
 * De-allocate object memory
 */
void bwsfree(const bwstring* s)
{
	if (s)
		sort_free(s);
}

/*
 * Copy content of src binary string to dst.
 * If the capacity of the dst string is not sufficient,
 * then the data is truncated.
 */
size_t bwscpy(bwstring *dst, const bwstring *src)
{
	size_t ret = src->len;
	size_t nums = ret;
	if (nums > dst->sz)
		nums = dst->sz;
	dst->len = nums;

	if (MB_CUR_MAX == 1) {
		memcpy(dst->data.cstr, src->data.cstr, nums);
		dst->data.cstr[dst->len] = '\0';
	} else {
		memcpy(dst->data.wstr, src->data.wstr, SIZEOF_WCHAR_STRING(nums+1));
		dst->data.wstr[dst->len] = L'\0';
	}

	return ret;
}

/*
 * Copy content of src binary string to dst,
 * with specified number of symbols to be copied.
 * If the capacity of the dst string is not sufficient,
 * then the data is truncated.
 */
bwstring* bwsncpy(bwstring *dst, const bwstring *src, size_t size)
{
	size_t nums = src->len;
	if (nums > dst->sz)
		nums = dst->sz;
	if (nums > size)
		nums = size;
	dst->len = nums;

	if (MB_CUR_MAX == 1) {
		memcpy(dst->data.cstr, src->data.cstr, nums);
		dst->data.cstr[dst->len] = '\0';
	} else {
		memcpy(dst->data.wstr, src->data.wstr, SIZEOF_WCHAR_STRING(nums+1));
		dst->data.wstr[dst->len] = L'\0';
	}

	return dst;
}

/*
 * Copy content of src binary string to dst,
 * with specified number of symbols to be copied.
 * An offset value can be specified, from the start of src string.
 * If the capacity of the dst string is not sufficient,
 * then the data is truncated.
 */
bwstring* bwsnocpy(bwstring *dst, const bwstring *src, size_t offset, size_t size)
{
	if (offset >= src->len) {
		if (dst->sz > 0) {
			dst->data.wstr[0] = 0;
			dst->len = 0;
		}
	} else {
		size_t nums = src->len - offset;
		if (nums > dst->sz)
			nums = dst->sz;
		if (nums > size)
			nums = size;
		dst->len = nums;
		if (MB_CUR_MAX == 1) {
			memcpy(dst->data.cstr, src->data.cstr + offset, (nums));
			dst->data.cstr[dst->len] = '\0';
		} else {
			memcpy(dst->data.wstr, src->data.wstr + offset, SIZEOF_WCHAR_STRING(nums));
			dst->data.wstr[dst->len] = L'\0';
		}
	}
	return dst;
}

/*
 * Compare two binary strings, according to the locale setting.
 * Ignore the difference between upper and lower case symbols.
 */
int bwscasecoll(const bwstring* bws1, const bwstring *bws2)
{
	if (MB_CUR_MAX == 1) {

		const size_t len = 128;
		static unsigned char s1[129];
		static unsigned char s2[129];

		unsigned char c1 = 0, c2 = 0;
		size_t i = 0;
		size_t maxlen = bws1->len;
		if (maxlen > bws2->len)
			maxlen = bws2->len;

		while (i < maxlen) {
			size_t j = 0;
			int res = 0;
			for (j = 0; ((j < len) && (i < maxlen)); ++j, ++i) {
				c1 = bws1->data.cstr[i];
				c2 = bws2->data.cstr[i];
				if (!c1 || !c2)
					break;
				s1[j] = tolower(c1);
				s2[j] = tolower(c2);
			}
			s1[j] = '\0';
			s2[j] = '\0';
			res = strcoll(s1, s2);
			if (res)
				return res;
			if (!c1 && c2)
				return -1;
			else if (!c2 && c1)
				return +1;
			else if (!c1 && !c2)
				++i;
		}

		if (bws1->len < bws2->len)
			return -1;
		else if (bws1->len > bws2->len)
			return +1;

	} else {

		const size_t len = 128;
		static wchar_t s1[129];
		static wchar_t s2[129];

		wchar_t c1 = 0, c2 = 0;
		size_t i = 0;
		size_t maxlen = bws1->len;
		if (maxlen > bws2->len)
			maxlen = bws2->len;

		while (i < maxlen) {
			size_t j = 0;
			int res = 0;
			for (j = 0; ((j < len) && (i < maxlen)); ++j, ++i) {
				c1 = bws1->data.wstr[i];
				c2 = bws2->data.wstr[i];
				if (!c1 || !c2)
					break;
				s1[j] = towlower(c1);
				s2[j] = towlower(c2);
			}
			s1[j] = L'\0';
			s2[j] = L'\0';
			res = wcscoll(s1, s2);
			if (res)
				return res;
			if (!c1 && c2)
				return -1;
			else if (!c2 && c1)
				return +1;
			else if (!c1 && !c2)
				++i;
		}

		if (bws1->len < bws2->len)
			return -1;
		else if (bws1->len > bws2->len)
			return +1;
	}

	return 0;
}

/*
 * Write binary string to the file.
 * The output is ended either with '\n' (nl == true)
 * or '\0' (nl == false).
 */
int bwsfwrite(bwstring* bws, FILE* f, bool zero_ended)
{
	if (MB_CUR_MAX == 1) {

		size_t len = bws->len;

		if (!zero_ended) {

			bws->data.cstr[len] = '\n';

			if (fwrite(bws->data.cstr, len + 1, 1, f) < 1)
				err(2, NULL);

			bws->data.cstr[len] = '\0';

		} else {

			if (fwrite(bws->data.cstr, len + 1, 1, f) < 1)
				err(2, NULL);

		}

		return len + 1;

	} else {

		wchar_t eols = zero_ended ? btowc('\0') : btowc('\n');
		int printed = 0;
		while (printed < (int) bwslen(bws)) {
			const wchar_t *s = bws->data.wstr + printed;
			if (*s == L'\0') {
				int nums = fwprintf(f, L"%lc", *s);
				if (nums != 1) {
					err(2, NULL);
				}
				++printed;
			} else {
				int nums = fwprintf(f, L"%ls", s);
				if (nums < 1) {
					err(2, NULL);
				}
				printed += nums;
			}
		}
		fwprintf(f, L"%lc", eols);
		return printed + 1;
	}
}

/*
 * Allocate and read a binary string from file.
 * The strings are nl-ended or zero-ended, depending on the sort setting.
 */
bwstring *bwsfgetln(FILE* f, size_t *len, bool zero_ended)
{
	const wchar_t eols = zero_ended ? btowc('\0') : btowc('\n');

	if (!zero_ended && (MB_CUR_MAX > 1)) {

		wchar_t *ret = fgetwln(f, len);
		if (!ret) {
			if (!feof(f)) {
				err(2, NULL);
			}
			return NULL;
		}
		if (*len > 0) {
			if (ret[*len - 1] == eols) {
				--(*len);
			}
		}
		return bwssbdup(ret, *len);

	} else {

		static wchar_t* fgetwln_z_buffer = NULL;
		static size_t fgetwln_z_buffer_size = 0;
		wchar_t c = 0;

		*len = 0;

		if (feof(f))
			return NULL;

		if (2 >= fgetwln_z_buffer_size) {
			fgetwln_z_buffer_size += 256;
			fgetwln_z_buffer = sort_realloc(fgetwln_z_buffer, sizeof(wchar_t) * fgetwln_z_buffer_size);
		}
		fgetwln_z_buffer[*len] = 0;

		if (MB_CUR_MAX == 1) {

			while (!feof(f)) {

				c = fgetc(f);

				if (c == EOF) {
					if (*len == 0)
						return NULL;
					goto line_read_done;
				}
				if (c == eols) {
					goto line_read_done;
				}

				if (*len + 1 >= fgetwln_z_buffer_size) {
					fgetwln_z_buffer_size += 256;
					fgetwln_z_buffer
									= sort_realloc(
													fgetwln_z_buffer,
													SIZEOF_WCHAR_STRING(fgetwln_z_buffer_size));
				}

				fgetwln_z_buffer[*len] = c;
				fgetwln_z_buffer[++(*len)] = 0;
			}

		} else {

			while (!feof(f)) {

				c = fgetwc(f);

				if (c == WEOF) {
					if (*len == 0)
						return NULL;
					goto line_read_done;
				}
				if (c == eols) {
					goto line_read_done;
				}

				if (*len + 1 >= fgetwln_z_buffer_size) {
					fgetwln_z_buffer_size += 256;
					fgetwln_z_buffer
									= sort_realloc(
													fgetwln_z_buffer,
													SIZEOF_WCHAR_STRING(fgetwln_z_buffer_size));
				}

				fgetwln_z_buffer[*len] = c;
				fgetwln_z_buffer[++(*len)] = 0;
			}
		}

		line_read_done:
		/* we do not count the last 0 */

		return bwssbdup(fgetwln_z_buffer, *len);
	}
}

int bwscoll(const bwstring* bws1, const bwstring *bws2)
{
	size_t len1 = bws1->len;
	size_t len2 = bws2->len;

	if (byte_sort && (MB_CUR_MAX == 1)) {
		int res = 0;
		if (len1 > len2) {
			res = memcmp(bws1->data.cstr, bws2->data.cstr, len2);
			if (!res)
				res = +1;
		} else if (len1 < len2) {
			res = memcmp(bws1->data.cstr, bws2->data.cstr, len1);
			if (!res)
				res = -1;
		} else {
			res = memcmp(bws1->data.cstr, bws2->data.cstr, len1);
		}
		return res;

	} else if (MB_CUR_MAX == 1) {

		const unsigned char* s1 = bws1->data.cstr;
		const unsigned char* s2 = bws2->data.cstr;
		int res = 0;
		size_t i = 0;
		size_t maxlen = len1;

		if (maxlen > len2)
			maxlen = len2;

		while (i < maxlen) {

			/* goto next non-zero part: */
			while ((i < maxlen) && !s1[i] && !s2[i])
				++i;

			if (i >= maxlen)
				break;

			if (s1[i] == 0) {
				if (s2[i] == 0) {
					/* must never come here */
					err(2, "bwscoll error 01");
				} else {
					return -1;
				}
			} else if (s2[i] == 0) {
				return +1;
			}

			res = strcoll(s1 + i, s2 + i);
			if (res)
				return res;

			while ((i < maxlen) && s1[i] && s2[i])
				++i;

			if (i >= maxlen)
				break;

			if (s1[i] == 0) {
				if (s2[i] == 0) {
					++i;
					continue;
				} else {
					return -1;
				}
			} else if (s2[i] == 0) {
				return +1;
			} else {
				/* must never be here */
				err(2, "bwscoll error 02");
			}
		}

		if (len1 < len2)
			return -1;
		else if (len1 > len2)
			return +1;

		return 0;

	} else {

		const wchar_t* s1 = bws1->data.wstr;
		const wchar_t* s2 = bws2->data.wstr;
		int res = 0;
		size_t i = 0;
		size_t maxlen = len1;

		if (maxlen > len2)
			maxlen = len2;

		while (i < maxlen) {

			/* goto next non-zero part: */
			while ((i < maxlen) && !s1[i] && !s2[i])
				++i;

			if (i >= maxlen)
				break;

			if (s1[i] == 0) {
				if (s2[i] == 0) {
					/* must never come here */
					err(2, "bwscoll error 1");
				} else {
					return -1;
				}
			} else if (s2[i] == 0) {
				return +1;
			}

			res = wcscoll(s1 + i, s2 + i);
			if (res)
				return res;

			while ((i < maxlen) && s1[i] && s2[i])
				++i;

			if (i >= maxlen)
				break;

			if (s1[i] == 0) {
				if (s2[i] == 0) {
					++i;
					continue;
				} else {
					return -1;
				}
			} else if (s2[i] == 0) {
				return +1;
			} else {
				/* must never be here */
				err(2, "bwscoll error 2");
			}
		}

		if (len1 < len2)
			return -1;
		else if (len1 > len2)
			return +1;

		return 0;
	}
}

/*
 * Correction of the system API
 */
double bwstod(bwstring *s0, bool *empty)
{
	double ret = 0;

	if (MB_CUR_MAX == 1) {
		unsigned char *s = s0->data.cstr;
		unsigned char *end = s + s0->len;
		char *ep = NULL;

		while (isblank(*s) && s < end)
			++s;

		if (!isprint(*s)) {
			*empty = true;
			return 0;
		}

		ret = strtod(s, &ep);
		if ((unsigned char*) ep == s) {
			*empty = true;
			return 0;
		}

	} else {

		wchar_t *s = s0->data.wstr;
		wchar_t *end = s + s0->len;
		wchar_t *ep = NULL;

		while (iswblank(*s) && s < end)
			++s;

		if (!iswprint(*s)) {
			*empty = true;
			return 0;
		}

		ret = wcstod(s, &ep);
		if (ep == s) {
			*empty = true;
			return 0;
		}
	}

	*empty = false;
	return ret;
}

/*
 * A helper function for monthcoll.  If a line matches
 * a month name, it returns (number of the month - 1),
 * while if there is no match, it just return -1.
 */

int bws_month_score(const bwstring *s0)
{
	if (MB_CUR_MAX == 1) {

		static unsigned char **cmonths = NULL;
		const unsigned char* s = s0->data.cstr;
		const unsigned char *end = s + s0->len;
		unsigned char *tmp;
		size_t len;

		while (isblank(*s) && s < end)
			++s;

		len = strlen(s);

		if (cmonths == NULL) {

			nl_item item[12] = { ABMON_1, ABMON_2, ABMON_3, ABMON_4, ABMON_5, ABMON_6, ABMON_7, ABMON_8,
					ABMON_9, ABMON_10, ABMON_11, ABMON_12 };

			unsigned char *m = NULL;
			int i = 0;

			cmonths = sort_malloc(sizeof(unsigned char*) * 12);
			for (i = 0; i < 12; i++) {
				unsigned int j = 0;
				cmonths[i] = NULL;
				tmp = (unsigned char*) nl_langinfo(item[i]);
				if (!tmp)
					continue;
				len = strlen(tmp);
				if (len < 1)
					continue;
				while (isblank(*tmp))
					++tmp;
				m = sort_malloc(len + 1);
				memcpy(m, tmp, len + 1);
				m[len] = '\0';
				for (j = 0; j < len; j++)
					m[j] = tolower(m[j]);
				cmonths[i] = m;
			}
		}

		{
			int i = 0;
			for (i = 11; i >= 0; --i) {
				if (cmonths[i] && (s == (unsigned char*) strstr(s, cmonths[i])))
					return (i);
			}
		}

	} else {

		static wchar_t **wmonths = NULL;
		const wchar_t* s = s0->data.wstr;
		const wchar_t *end = s + s0->len;
		unsigned char *tmp;
		size_t len;

		while (iswblank(*s) && s < end)
			++s;

		len = wcslen(s);

		if (wmonths == NULL) {

			nl_item item[12] = { ABMON_1, ABMON_2, ABMON_3, ABMON_4, ABMON_5, ABMON_6, ABMON_7, ABMON_8,
					ABMON_9, ABMON_10, ABMON_11, ABMON_12 };

			wchar_t *m;
			int i = 0;

			wmonths = sort_malloc(sizeof(wchar_t *) * 12);
			for (i = 0; i < 12; i++) {
				unsigned int j = 0;
				wmonths[i] = NULL;
				tmp = (unsigned char*) nl_langinfo(item[i]);
				if (!tmp)
					continue;
				len = strlen(tmp);
				if (len < 1)
					continue;
				while (isblank(*tmp))
					++tmp;
				m = sort_malloc(SIZEOF_WCHAR_STRING(len + 1));
				if (mbstowcs(m, tmp, len) == ((size_t) -1))
					continue;
				m[len] = L'\0';
				for (j = 0; j < len; j++)
					m[j] = towlower(m[j]);
				wmonths[i] = m;
			}
		}

		{
			int i = 0;
			for (i = 11; i >= 0; --i) {
				if (wmonths[i] && (s == wcsstr(s, wmonths[i])))
					return (i);
			}
		}

	}

	return (-1);
}

/*
 * Rips out leading blanks (-b).
 */
bwstring *ignore_leading_blanks(bwstring *str)
{
	if (MB_CUR_MAX == 1) {

		unsigned char *src = str->data.cstr;
		unsigned char *dst = src;
		unsigned char *end = src + str->len;

		while (src < end && isblank(*src))
			++src;

		if (src != dst) {
			size_t newlen = bwslen(str) - (src - dst);
			while (src < end) {
				*dst = *src;
				++dst;
				++src;
			}
			bws_setlen(str, newlen);

		}

	} else {

		wchar_t *src = str->data.wstr;
		wchar_t *dst = src;
		wchar_t *end = src + str->len;

		while (src < end && iswblank(*src))
			++src;

		if (src != dst) {
			size_t newlen = bwslen(str) - (src - dst);
			while (src < end) {
				*dst = *src;
				++dst;
				++src;
			}
			bws_setlen(str, newlen);

		}
	}

	return str;
}

/*
 * Rips out nonprinting characters (-i).
 */
bwstring *ignore_nonprinting(bwstring *str)
{
	size_t newlen = str->len;

	if (MB_CUR_MAX == 1) {

		unsigned char *src = str->data.cstr;
		unsigned char *dst = src;
		unsigned char *end = src + str->len;
		unsigned char c = 0;

		while (src < end) {
			c = *src;
			if (isprint(c)) {
				*dst = c;
				++dst;
				++src;
			} else {
				++src;
				--newlen;
			}
		}

	} else {

		wchar_t *src = str->data.wstr;
		wchar_t *dst = src;
		wchar_t *end = src + str->len;
		wchar_t c = 0;

		while (src < end) {
			c = *src;
			if (iswprint(c)) {
				*dst = c;
				++dst;
				++src;
			} else {
				++src;
				--newlen;
			}
		}
	}

	bws_setlen(str, newlen);

	return str;
}

/*
 * Rips out any characters that are not alphanumeric characters
 * nor blanks (-d).
 */
bwstring *dictionary_order(bwstring *str)
{
	size_t newlen = str->len;

	if (MB_CUR_MAX == 1) {

		unsigned char *src = str->data.cstr;
		unsigned char *dst = src;
		unsigned char *end = src + str->len;
		unsigned char c = 0;

		while (src < end) {
			c = *src;
			if (isalnum(c) || isblank(c)) {
				*dst = c;
				++dst;
				++src;
			} else {
				++src;
				--newlen;
			}
		}

	} else {

		wchar_t *src = str->data.wstr;
		wchar_t *dst = src;
		wchar_t *end = src + str->len;
		wchar_t c = 0;

		while (src < end) {
			c = *src;
			if (iswalnum(c) || iswblank(c)) {
				*dst = c;
				++dst;
				++src;
			} else {
				++src;
				--newlen;
			}
		}
	}

	bws_setlen(str, newlen);

	return str;
}

/*
 * Converts string to lower case(-f).
 */
bwstring *ignore_case(bwstring *str)
{
	if (MB_CUR_MAX == 1) {

		unsigned char *s = str->data.cstr;
		unsigned char *end = s + str->len;

		while (s < end) {
			*s = tolower(*s);
			++s;
		}

	} else {
		wchar_t *s = str->data.wstr;
		wchar_t *end = s + str->len;

		while (s < end) {
			*s = towlower(*s);
			++s;
		}
	}

	return str;
}

void bws_disorder_warnx(bwstring* s, const char* fn, size_t pos)
{
	if (MB_CUR_MAX == 1) {
		warnx("%s:%d: disorder: %s", fn, (int) (pos + 1), s->data.cstr);
	} else {
		warnx("%s:%d: disorder: %ls", fn, (int) (pos + 1), s->data.wstr);
	}
}
