/* Copyright (C) 2012 Oleg Moskalenko oleg.moskalenko@citrix.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/limits.h>

#include <errno.h>
#include <err.h>
#include <langinfo.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <wctype.h>

#include "radixsort.h"
#include "coll.h"

static unsigned char *ctable=NULL;
static unsigned char *indtable=NULL;

typedef struct _sort_level {

	size_t level;

	sort_list_item** tosort;
	size_t tosort_num;
	size_t tosort_sz;

	sort_list_item** leaves;
	size_t leaves_num;
	size_t leaves_sz;

	struct _sort_level** sublevels;
	size_t sln;
	size_t real_sln;
} sort_level;

static int cmpwchar(const void *pc1, const void *pc2)
{
	int res=0;
	const unsigned char *c1=(const unsigned char*)pc1;
	const unsigned char *c2=(const unsigned char*)pc2;
	wchar_t s1[2];
	wchar_t s2[2];
	s1[0]=*c1;
	s1[1]=L'\0';
	s2[0]=*c2;
	s2[1]=L'\0';

	res = wcscoll(s1,s2);

	return res;
}

static void make_char_table(void)
{
	if(!ctable) {
		int i=0;
		ctable=(unsigned char*)sort_malloc(256);
		for(i=0;i<256;++i) {
			ctable[i]=i;
		}
		qsort(ctable,256,1,cmpwchar);

		indtable=(unsigned char*)sort_malloc(256);
		memset(indtable,0,256);
		for(i=0;i<256;++i) {
			unsigned char c = ctable[i];
			unsigned char pos = (unsigned char)i;
			indtable[c] = pos;
			/* printf("%s: c=%d, pos=%d\n",__FUNCTION__,(int)c,(int)pos); */
		}
	}
}

static void add_to_sublevel(sort_level* sl, sort_list_item *item, int indx)
{
	sort_level* ssl = sl->sublevels[indx];
	if(!ssl) {
		ssl=(sort_level*)sort_malloc(sizeof(sort_level));
		if(!ssl)
			err(2,NULL);
		memset(ssl,0,sizeof(sort_level));

		ssl->level=sl->level+1;
		sl->sublevels[indx] = ssl;

		++(sl->real_sln);
	}

	if(++(ssl->tosort_num)>ssl->tosort_sz) {
		ssl->tosort_sz=ssl->tosort_num+128;
		ssl->tosort=(sort_list_item**)sort_realloc(ssl->tosort,sizeof(sort_list_item*)*(ssl->tosort_sz));
	}

	ssl->tosort[ssl->tosort_num-1]=item;
}

static inline void add_leaf(sort_level* sl, sort_list_item *item) {
	if(++(sl->leaves_num) > sl->leaves_sz) {
		sl->leaves_sz=sl->leaves_num+128;
		sl->leaves=(sort_list_item**)sort_realloc(sl->leaves,(sizeof(sort_list_item*)*(sl->leaves_sz)));
	}
	sl->leaves[sl->leaves_num-1]=item;
}

static inline int get_wc_index(sort_list_item* sli, size_t level) {
	const bwstring *bws=sli->ks->set[0].k;
	if((bwslen(bws)>level)) {
		return indtable[(unsigned char)bws_get(bws,level)];
	}
	return -1;
}

static void place_item(sort_level* sl, size_t item)
{
	sort_list_item* sli=sl->tosort[item];
	int c = get_wc_index(sli,sl->level);
	if(c == -1) {
		add_leaf(sl,sli);
	} else {
		add_to_sublevel(sl,sli,c);
	}
}

static const size_t slsz=256*sizeof(sort_level*);
static sort_level* tmp_sublevels[256];

static void free_sort_level(sort_level *sl)
{
	if(sl) {
		size_t i=0;
		if(sl->leaves) sort_free(sl->leaves);
		if(sl->level>0 && sl->tosort) sort_free(sl->tosort);
		if(sl->sublevels) {
			for(i=0;i<sl->sln;++i)
				free_sort_level(sl->sublevels[i]);
			sort_free(sl->sublevels);
		}
		sort_free(sl);
	}
}

static void run_sort_level_next(sort_level* sl, sort_list_item** sorted, size_t *sorted_n)
{
	size_t i = 0;
	size_t tosort_num=0;
	size_t sln=0;

	if(sl->sublevels) {
		sort_free(sl->sublevels);
		sl->sublevels=NULL;
	}

	switch (sl->tosort_num){
	case 0:
		if(sl->tosort) {
			sort_free(sl->tosort);
			sl->tosort=NULL;
		}
		return;
	case (1):
		sorted[(*sorted_n)++]=sl->tosort[0];
		sort_free(sl->tosort);
		sl->tosort = NULL;
		return;
	case (2):
		if (list_coll(&(sl->tosort[0]), &(sl->tosort[1])) > 0) {
			sorted[(*sorted_n)++]=sl->tosort[1];
			sorted[(*sorted_n)++]=sl->tosort[0];
		} else {
			sorted[(*sorted_n)++]=sl->tosort[0];
			sorted[(*sorted_n)++]=sl->tosort[1];
		}
		sort_free(sl->tosort);
		sl->tosort=NULL;
		return;
	default:
		if ((sl->tosort_num < 65) || (sl->level > 15)) {
			sl->leaves = sl->tosort;
			sl->leaves_num = sl->tosort_num;
			sl->leaves_sz = sl->leaves_num;
			sl->leaves = (sort_list_item**) sort_realloc(sl->leaves, (sizeof(sort_list_item*)
							* (sl->leaves_sz)));
			sl->tosort = NULL;
			sl->tosort_num = 0;
			sl->tosort_sz = 0;
			sl->sln = 0;
			sl->real_sln = 0;
			if (sort_opts_vals.sflag) {
				if (mergesort(sl->leaves, sl->leaves_num, sizeof(sort_list_item *), (int(*)(
								const void *, const void *)) list_coll) == -1)
					err(2, "Radix sort error 3");
			} else {
				qsort(sl->leaves, sl->leaves_num, sizeof(sort_list_item *), (int(*)(const void *,
								const void *)) list_coll);
			}

			memcpy(sorted + (*sorted_n), sl->leaves, sl->leaves_num*sizeof(sort_list_item*));
			(*sorted_n) += sl->leaves_num;

			sort_free(sl->leaves);
			sl->leaves=NULL;
			sl->leaves_num=0;
			sl->leaves_sz=0;

			return;
		} else {
			sl->tosort_sz = sl->tosort_num;
			sl->tosort = (sort_list_item**) sort_realloc(sl->tosort, sizeof(sort_list_item*)
							* (sl->tosort_sz));
		}
	}

	sl->sln = 256;
	sl->sublevels = tmp_sublevels;
	memset(sl->sublevels, 0, slsz);

	sl->real_sln = 0;

	tosort_num=sl->tosort_num;
	for (i = 0; i < tosort_num; ++i) {
		place_item(sl, i);
	}

	sort_free(sl->tosort);
	sl->tosort = NULL;
	sl->tosort_num = 0;
	sl->tosort_sz = 0;

	{
		size_t new_sln = sl->real_sln;
		size_t newsz = 0;
		size_t counter = 0;
		sort_level** new_sls = NULL;

		if(new_sln>0) {
			newsz = new_sln * sizeof(sort_level*);
			new_sls = (sort_level**) sort_malloc(newsz);
			sln=sl->sln;
			for (i = 0; i < sln; ++i) {
				if (sl->sublevels[i]) {
					new_sls[counter++] = sl->sublevels[i];
				}
			}
		}

		sl->sublevels = new_sls;
		sl->sln = new_sln;
	}

	if (sl->leaves_num > 1) {
		if (!(sort_opts_vals.sflag) && sort_opts_vals.complex_sort) {
			qsort(sl->leaves, sl->leaves_num, sizeof(sort_list_item *),
						(int(*)(const void *, const void *)) list_coll_by_str_only);
		}
	}

	sl->leaves_sz = sl->leaves_num;
	sl->leaves=(sort_list_item**)sort_realloc(sl->leaves,(sizeof(sort_list_item*)*(sl->leaves_sz)));

	if (!(current_sort_mods->rflag)) {

		memcpy(sorted + (*sorted_n), sl->leaves, sl->leaves_num*sizeof(sort_list_item*));
		(*sorted_n) += sl->leaves_num;

		sort_free(sl->leaves);
		sl->leaves=NULL;
		sl->leaves_num=0;
		sl->leaves_sz=0;

		sln=sl->sln;
		for (i = 0; i < sln; ++i) {
			sort_level* slc=sl->sublevels[i];
			if(slc) {
				run_sort_level_next(slc,sorted,sorted_n);
				free_sort_level(slc);
				sl->sublevels[i]=NULL;
			}
		}

	} else {

		sln=sl->sln;
		for (i = 0; i < sln; ++i) {
			size_t n = sln - i - 1;
			sort_level* slc=sl->sublevels[n];
			if(slc) {
				run_sort_level_next(slc,sorted,sorted_n);
				free_sort_level(slc);
				sl->sublevels[n]=NULL;
			}
		}

		memcpy(sorted + (*sorted_n), sl->leaves, sl->leaves_num*sizeof(sort_list_item*));
		(*sorted_n) += sl->leaves_num;

		sort_free(sl->leaves);
		sl->leaves=NULL;
		sl->leaves_num=0;
		sl->leaves_sz=0;
	}


	if(sl->sublevels) {
		sort_free(sl->sublevels);
		sl->sublevels=NULL;
		sl->sln=0;
	}

}

static void run_top_sort_level(sort_level* sl)
{
	size_t i = 0;
	size_t sorted_n=0;

	sl->sln = 256;
	sl->sublevels = (sort_level**) sort_malloc(slsz);
	memset(sl->sublevels, 0, slsz);

	for (i = 0; i < sl->tosort_num; ++i) {
		place_item(sl, i);
	}

	if (sl->leaves_num > 1) {
		if (!(sort_opts_vals.sflag) && sort_opts_vals.complex_sort) {
			qsort(sl->leaves, sl->leaves_num, sizeof(sort_list_item *),
						(int(*)(const void *, const void *)) list_coll_by_str_only);
		}
	}

	if (!(current_sort_mods->rflag)) {
		memcpy(sl->tosort + sorted_n, sl->leaves, sl->leaves_num*sizeof(sort_list_item*));
		sorted_n += sl->leaves_num;
		for (i = 0; i < sl->sln; ++i) {
			sort_level* slc=sl->sublevels[i];
			if(slc) {
				run_sort_level_next(slc,sl->tosort,&sorted_n);
				free_sort_level(slc);
				sl->sublevels[i]=NULL;
			}
		}
	} else {
		for (i = 0; i < sl->sln; ++i) {
			size_t n = sl->sln - i - 1;
			sort_level* slc=sl->sublevels[n];
			if(slc) {
				run_sort_level_next(slc,sl->tosort,&sorted_n);
				free_sort_level(slc);
				sl->sublevels[n]=NULL;
			}
		}
		memcpy(sl->tosort + sorted_n, sl->leaves, sl->leaves_num*sizeof(sort_list_item*));
		sorted_n += sl->leaves_num;
	}
}

static void run_sort(sort_list_item **base, size_t nmemb)
{
	sort_level *sl=(sort_level*)sort_malloc(sizeof(sort_level));
	if(!sl)
		err(2,NULL);
	memset(sl,0,sizeof(sort_level));

	sl->tosort=base;
	sl->tosort_num=nmemb;
	sl->tosort_sz=nmemb;

	run_top_sort_level(sl);

	free_sort_level(sl);
}

void rxsort(sort_list_item **base, size_t nmemb)
{
	make_char_table();
	run_sort(base,nmemb);
}
